import yaml
import gdown
import zipfile
import os


def unzip(src: str, dst: str = ""):
    """unzip the file

    Args:
        src (str):
        dst (str):

    Returns:
        None
    """
    with zipfile.ZipFile(src, 'r') as zip_ref:
        if len(dst):
            zip_ref.extractall(dst)
        else:
            zip_ref.extractall()


def get_cfg(path: str):
    """Set up the config

    Args:
        path (str):

    Returns:
        None
    """
    
    yaml.warnings({'YAMLLoadWarning': False})
    with open(path, 'r') as fs:
        config = yaml.load(fs)
    return config


def download_url(id: str, name: str, dst: str = ""):
    """Downloads file from a url to a destination.
    
    Args:
        id (str): url to download file.
        dst (str): destination path.
    """

    url = "https://drive.google.com/uc?id=" + id
    output = name + '.zip'
    gdown.download(url, output, quiet=False)

    # unzip the file
    unzip(output, dst)

    # clean up space
    os.remove(output)



import numpy as np

from os.path import isfile


class Label:

	def __init__(self,cl=-1,tl=np.array([0.,0.]),br=np.array([0.,0.]),prob=None):
		self.__tl 	= tl
		self.__br 	= br
		self.__cl 	= cl
		self.__prob = prob

	def __str__(self):
		return 'Class: %d, top_left(x:%f,y:%f), bottom_right(x:%f,y:%f)' % (self.__cl, self.__tl[0], self.__tl[1], self.__br[0], self.__br[1])

	def copy(self):
		return Label(self.__cl,self.__tl,self.__br)

	def wh(self): return self.__br-self.__tl

	def cc(self): return self.__tl + self.wh()/2

	def tl(self): return self.__tl
 
	def br(self): return self.__br

	def tr(self): return np.array([self.__br[0],self.__tl[1]])

	def bl(self): return np.array([self.__tl[0],self.__br[1]])

	def cl(self): return self.__cl

	def area(self): return np.prod(self.wh())

	def prob(self): return self.__prob

	def set_class(self,cl):
		self.__cl = cl

	def set_tl(self,tl):
		self.__tl = tl

	def set_br(self,br):
		self.__br = br

	def set_wh(self,wh):
		cc = self.cc()
		self.__tl = cc - .5*wh
		self.__br = cc + .5*wh

	def set_prob(self,prob):
		self.__prob = prob


def lread(file_path,label_type=Label):

	if not isfile(file_path):
		return []

	objs = []
	with open(file_path,'r') as fd:
		for line in fd:
			v 		= line.strip().split()
			cl 		= int(v[0])
			ccx,ccy = float(v[1]),float(v[2])
			w,h 	= float(v[3]),float(v[4])
			prob 	= float(v[5]) if len(v) == 6 else None

			cc 	= np.array([ccx,ccy])
			wh 	= np.array([w,h])

			objs.append(label_type(cl,cc-wh/2,cc+wh/2,prob=prob))

	return objs

def lwrite(file_path,labels,write_probs=True):
	with open(file_path,'w') as fd:
		for l in labels:
			cc,wh,cl,prob = (l.cc(),l.wh(),l.cl(),l.prob())
			if prob != None and write_probs:
				fd.write('%d %f %f %f %f %f\n' % (cl,cc[0],cc[1],wh[0],wh[1],prob))
			else:
				fd.write('%d %f %f %f %f\n' % (cl,cc[0],cc[1],wh[0],wh[1]))


def dknet_label_conversion(R,img_width,img_height):
	WH = np.array([img_width,img_height],dtype=float)
	L  = []
	for r in R:
		center = np.array(r[2][:2])/WH
		wh2 = (np.array(r[2][2:])/WH)*.5
		L.append(Label(ord(r[0]),tl=center-wh2,br=center+wh2,prob=r[1]))
	return L


class Shape():

	def __init__(self,pts=np.zeros((2,0)),max_sides=4,text=''):
		self.pts = pts
		self.max_sides = max_sides
		self.text = text

	def isValid(self):
		return self.pts.shape[1] > 2

	def write(self,fp):
		fp.write('%d,' % self.pts.shape[1])
		ptsarray = self.pts.flatten()
		fp.write(''.join([('%f,' % value) for value in ptsarray]))
		fp.write('%s,' % self.text)
		fp.write('\n')

	def read(self,line):
		data 		= line.strip().split(',')
		ss 			= int(data[0])
		values 		= data[1:(ss*2 + 1)]
		text 		= data[(ss*2 + 1)] if len(data) >= (ss*2 + 2) else ''
		self.pts 	= np.array([float(value) for value in values]).reshape((2,ss))
		self.text   = text

def readShapes(path,obj_type=Shape):
	shapes = []
	with open(path) as fp:
		for line in fp:
			shape = obj_type()
			shape.read(line)
			shapes.append(shape)
	return shapes

def writeShapes(path,shapes):
	if len(shapes):
		with open(path,'w') as fp:
			for shape in shapes:
				if shape.isValid():
					shape.write(fp)


# author:    Adrian Rosebrock
# website:   http://www.pyimagesearch.com

# import the necessary packages
import numpy as np
import cv2
import sys

# import any special Python 2.7 packages
if sys.version_info.major == 2:
    from urllib import urlopen

# import any special Python 3 packages
elif sys.version_info.major == 3:
    from urllib.request import urlopen

def translate(image, x, y):
    # define the translation matrix and perform the translation
    M = np.float32([[1, 0, x], [0, 1, y]])
    shifted = cv2.warpAffine(image, M, (image.shape[1], image.shape[0]))

    # return the translated image
    return shifted

def rotate(image, angle, center=None, scale=1.0):
    # grab the dimensions of the image
    (h, w) = image.shape[:2]

    # if the center is None, initialize it as the center of
    # the image
    if center is None:
        center = (w // 2, h // 2)

    # perform the rotation
    M = cv2.getRotationMatrix2D(center, angle, scale)
    rotated = cv2.warpAffine(image, M, (w, h))

    # return the rotated image
    return rotated

def rotate_bound(image, angle):
    # grab the dimensions of the image and then determine the
    # center
    (h, w) = image.shape[:2]
    (cX, cY) = (w / 2, h / 2)

    # grab the rotation matrix (applying the negative of the
    # angle to rotate clockwise), then grab the sine and cosine
    # (i.e., the rotation components of the matrix)
    M = cv2.getRotationMatrix2D((cX, cY), -angle, 1.0)
    cos = np.abs(M[0, 0])
    sin = np.abs(M[0, 1])

    # compute the new bounding dimensions of the image
    nW = int((h * sin) + (w * cos))
    nH = int((h * cos) + (w * sin))

    # adjust the rotation matrix to take into account translation
    M[0, 2] += (nW / 2) - cX
    M[1, 2] += (nH / 2) - cY

    # perform the actual rotation and return the image
    return cv2.warpAffine(image, M, (nW, nH))

def resize(image, width=None, height=None, inter=cv2.INTER_AREA):
    # initialize the dimensions of the image to be resized and
    # grab the image size
    dim = None
    (h, w) = image.shape[:2]

    # if both the width and height are None, then return the
    # original image
    if width is None and height is None:
        return image

    # check to see if the width is None
    if width is None:
        # calculate the ratio of the height and construct the
        # dimensions
        r = height / float(h)
        dim = (int(w * r), height)

    # otherwise, the height is None
    else:
        # calculate the ratio of the width and construct the
        # dimensions
        r = width / float(w)
        dim = (width, int(h * r))

    # resize the image
    resized = cv2.resize(image, dim, interpolation=inter)

    # return the resized image
    return resized

def skeletonize(image, size, structuring=cv2.MORPH_RECT):
    # determine the area (i.e. total number of pixels in the image),
    # initialize the output skeletonized image, and construct the
    # morphological structuring element
    area = image.shape[0] * image.shape[1]
    skeleton = np.zeros(image.shape, dtype="uint8")
    elem = cv2.getStructuringElement(structuring, size)

    # keep looping until the erosions remove all pixels from the
    # image
    while True:
        # erode and dilate the image using the structuring element
        eroded = cv2.erode(image, elem)
        temp = cv2.dilate(eroded, elem)

        # subtract the temporary image from the original, eroded
        # image, then take the bitwise 'or' between the skeleton
        # and the temporary image
        temp = cv2.subtract(image, temp)
        skeleton = cv2.bitwise_or(skeleton, temp)
        image = eroded.copy()

        # if there are no more 'white' pixels in the image, then
        # break from the loop
        if area == area - cv2.countNonZero(image):
            break

    # return the skeletonized image
    return skeleton

def opencv2matplotlib(image):
    # OpenCV represents images in BGR order; however, Matplotlib
    # expects the image in RGB order, so simply convert from BGR
    # to RGB and return
    return cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

def url_to_image(url, readFlag=cv2.IMREAD_COLOR):
    # download the image, convert it to a NumPy array, and then read
    # it into OpenCV format
    resp = urlopen(url)
    image = np.asarray(bytearray(resp.read()), dtype="uint8")
    image = cv2.imdecode(image, readFlag)

    # return the image
    return image

def auto_canny(image, sigma=0.33):
    # compute the median of the single channel pixel intensities
    v = np.median(image)

    # apply automatic Canny edge detection using the computed median
    lower = int(max(0, (1.0 - sigma) * v))
    upper = int(min(255, (1.0 + sigma) * v))
    edged = cv2.Canny(image, lower, upper)

    # return the edged image
    return edged

def grab_contours(cnts):
    # if the length the contours tuple returned by cv2.findContours
    # is '2' then we are using either OpenCV v2.4, v4-beta, or
    # v4-official
    if len(cnts) == 2:
        cnts = cnts[0]

    # if the length of the contours tuple is '3' then we are using
    # either OpenCV v3, v4-pre, or v4-alpha
    elif len(cnts) == 3:
        cnts = cnts[1]

    # otherwise OpenCV has changed their cv2.findContours return
    # signature yet again and I have no idea WTH is going on
    else:
        raise Exception(("Contours tuple must have length 2 or 3, "
            "otherwise OpenCV changed their cv2.findContours return "
            "signature yet again. Refer to OpenCV's documentation "
            "in that case"))

    # return the actual contours array
    return cnts

def is_cv2(or_better=False):
    # grab the OpenCV major version number
    major = get_opencv_major_version()

    # check to see if we are using *at least* OpenCV 2
    if or_better:
        return major >= 2

    # otherwise we want to check for *strictly* OpenCV 2
    return major == 2

def is_cv3(or_better=False):
    # grab the OpenCV major version number
    major = get_opencv_major_version()

    # check to see if we are using *at least* OpenCV 3
    if or_better:
        return major >= 3

    # otherwise we want to check for *strictly* OpenCV 3
    return major == 3

def is_cv4(or_better=False):
    # grab the OpenCV major version number
    major = get_opencv_major_version()

    # check to see if we are using *at least* OpenCV 4
    if or_better:
        return major >= 4

    # otherwise we want to check for *strictly* OpenCV 4
    return major == 4

def get_opencv_major_version(lib=None):
    # if the supplied library is None, import OpenCV
    if lib is None:
        import cv2 as lib

    # return the major version number
    return int(lib.__version__.split(".")[0])

def check_opencv_version(major, lib=None):
    # this function may be removed in a future release as we now
    # use the get_opencv_major_function to obtain the current OpenCV
    # version and then perform the actual version check *within* the
    # respective function
    import warnings
    message = """
        The check_opencv_version function is deprecated and may be
        removed in a future release. Use at your own risk.
    """
    warnings.warn(message, DeprecationWarning, stacklevel=2)
    
    # if the supplied library is None, import OpenCV
    if lib is None:
        import cv2 as lib
        
    # return whether or not the current OpenCV version matches the
    # major version number
    return lib.__version__.startswith(major)

def build_montages(image_list, image_shape, montage_shape):
    """
    ---------------------------------------------------------------------------------------------
    author: Kyle Hounslow
    ---------------------------------------------------------------------------------------------
    Converts a list of single images into a list of 'montage' images of specified rows and columns.
    A new montage image is started once rows and columns of montage image is filled.
    Empty space of incomplete montage images are filled with black pixels
    ---------------------------------------------------------------------------------------------
    :param image_list: python list of input images
    :param image_shape: tuple, size each image will be resized to for display (width, height)
    :param montage_shape: tuple, shape of image montage (width, height)
    :return: list of montage images in numpy array format
    ---------------------------------------------------------------------------------------------

    example usage:

    # load single image
    img = cv2.imread('lena.jpg')
    # duplicate image 25 times
    num_imgs = 25
    img_list = []
    for i in xrange(num_imgs):
        img_list.append(img)
    # convert image list into a montage of 256x256 images tiled in a 5x5 montage
    montages = make_montages_of_images(img_list, (256, 256), (5, 5))
    # iterate through montages and display
    for montage in montages:
        cv2.imshow('montage image', montage)
        cv2.waitKey(0)

    ----------------------------------------------------------------------------------------------
    """
    if len(image_shape) != 2:
        raise Exception('image shape must be list or tuple of length 2 (rows, cols)')
    if len(montage_shape) != 2:
        raise Exception('montage shape must be list or tuple of length 2 (rows, cols)')
    image_montages = []
    # start with black canvas to draw images onto
    montage_image = np.zeros(shape=(image_shape[1] * (montage_shape[1]), image_shape[0] * montage_shape[0], 3),
                          dtype=np.uint8)
    cursor_pos = [0, 0]
    start_new_img = False
    for img in image_list:
        if type(img).__module__ != np.__name__:
            raise Exception('input of type {} is not a valid numpy array'.format(type(img)))
        start_new_img = False
        img = cv2.resize(img, image_shape)
        # draw image to black canvas
        montage_image[cursor_pos[1]:cursor_pos[1] + image_shape[1], cursor_pos[0]:cursor_pos[0] + image_shape[0]] = img
        cursor_pos[0] += image_shape[0]  # increment cursor x position
        if cursor_pos[0] >= montage_shape[0] * image_shape[0]:
            cursor_pos[1] += image_shape[1]  # increment cursor y position
            cursor_pos[0] = 0
            if cursor_pos[1] >= montage_shape[1] * image_shape[1]:
                cursor_pos = [0, 0]
                image_montages.append(montage_image)
                # reset black canvas
                montage_image = np.zeros(shape=(image_shape[1] * (montage_shape[1]), image_shape[0] * montage_shape[0], 3),
                                      dtype=np.uint8)
                start_new_img = True
    if start_new_img is False:
        image_montages.append(montage_image)  # add unfinished montage
    return image_montages


def adjust_brightness_contrast(image, brightness=0., contrast=0.):
    """
    Adjust the brightness and/or contrast of an image

    :param image: OpenCV BGR image
    :param contrast: Float, contrast adjustment with 0 meaning no change
    :param brightness: Float, brightness adjustment with 0 meaning no change
    """
    beta = 0
    # See the OpenCV docs for more info on the `beta` parameter to addWeighted
    # https://docs.opencv.org/3.4.2/d2/de8/group__core__array.html#gafafb2513349db3bcff51f54ee5592a19
    return cv2.addWeighted(image,
                           1 + float(contrast) / 100.,
                           image,
                           beta,
                           float(brightness))


from __future__ import absolute_import

import os
import sys

from .sanet import SANet
from .edet import EfficientDet

from crowd_counting.utils.tools import get_cfg, download_model, download_url

__model_factory = {
    # density map-based counting
    'densitymap': SANet,
    # detection-based counting
    'detection': EfficientDet
}

_model_urls = {
    'densitymap': ['1Gfbxk1ONNkdp_aPLnaYftbWpz7D97BHN',
                    '1BT1O0JghB80RnYrhumicBbxORKGX-wpO'],
    'detection': '1LItLb0V5lHMJM-LYx4uWEFe_NbcIm0Gf'
}

def show_avai_models():
    """Displays available models.

    Examples::
        >>> from crowd_counting import models
        >>> models.show_avai_models()
    """
    print(list(__model_factory.keys()))


def build_model(name: str, cfg_path: str = ""):
    """A function wrapper for building a model.

    Args:
        name (str): model name
        cfg_path (str): path to config file
    Returns:
        model instance.
    """
    avai_models = list(__model_factory.keys())
    if name not in avai_models:
        raise KeyError(
            'Unknown model: {}. Must be one of {}'.format(name, avai_models)
        )
    if not len(cfg_path):
        raise RuntimeError(
            'Please input the path config file'
        )
    # Load config file
    config = get_cfg(cfg_path)

    try:
        os.mkdir('models')
    except:
        pass
    
    try:
        os.mkdir('models/CrowdCounting')
    except:
        pass

    if name=='densitymap':
        # Check the models are existed or not
        if not os.path.exists('models/CrowdCounting/densitymap.hdf5'):
            download_model(_model_urls['densitymap'][0], name='densitymap', ext='.hdf5', dst = "models/CrowdCounting")
        if not os.path.exists('models/CrowdCounting/densitymap.json'):
            download_model(_model_urls['densitymap'][1], name='densitymap', ext='.json', dst = "models/CrowdCounting")
    elif name=='detection':
        # Check the models are existed or not
        if not os.path.exists('models/CrowdCounting/detection.h5'):
            download_model(_model_urls['detection'], name='detection.h5', dst = "models/CrowdCounting")

    return __model_factory[name](config)



import os
import tensorflow as tf
from tensorflow.keras.models import model_from_json
import numpy as np


class SANet():
    """ SANet Model
    """

    def __init__(self, config):
        self.model = None
        self.config = config

        self.load_model()

    def load_model(self):
        """Load crowd counting model"""

        model_path = self.config['model1']['path']
        json_path = self.config['model1']['cfg_json']
        gpu_device = self.config['gpu']

        if gpu_device==-1:
            with tf.device('/cpu:0'):
                with open(json_path, 'r') as json_file:
                    model_json = json_file.read()

                self.model = model_from_json(model_json)
                self.model.load_weights(model_path)
                print('Load the pretrained model sucessfully!')
        else:
            with tf.device('/gpu:{}'.format(gpu_device)):
                with open(json_path, 'r') as json_file:
                    model_json = json_file.read()

                self.model = model_from_json(model_json)
                self.model.load_weights(model_path)
                print('Load the pretrained model sucessfully in GPU mode!')


    def get_densitymap(self, image):
        """ Get the density map from the input image

        Args:
            image (np.array): the RGB image

        Returns:

        """
        try:
            output = self.model.predict(np.expand_dims(image, axis=0))
            output = np.squeeze(output)
            return output
        except:
            return None



# author:	Hung Le Viet

# set the version number
__version__ = "0.1.0"

# import the necessary packages
from .convenience import translate
from .convenience import rotate
from .convenience import rotate_bound
from .convenience import resize
from .convenience import skeletonize
from .convenience import opencv2matplotlib
from .convenience import url_to_image
from .convenience import auto_canny
from .convenience import grab_contours
from .convenience import is_cv2
from .convenience import is_cv3
from .convenience import is_cv4
from .convenience import check_opencv_version
from .convenience import build_montages
from .convenience import adjust_brightness_contrast
from .meta import find_function



from __future__ import division, absolute_import

import os
import sys
import time
import cv2
import mxnet as mx
import numpy as np
from mxnet import ndarray as nd
import datetime

from face_recognition.utils.tools import download_url

try:
    from rcnn.processing.bbox_transform import clip_boxes
    from rcnn.processing.generate_anchor import generate_anchors_fpn, anchors_plane
    from rcnn.processing.nms import gpu_nms_wrapper, cpu_nms_wrapper
except:
    print("[INFO] Not found rcnn library")
    print("[INFO] Start to download library")

    if not os.path.exists("rcnn"):
        download_url("11uZoI6ebGa4cEhaM9RuFO7yKiLE8mSCG", name="rcnn")

    download_url("1Iha2C-yH19XDA1LixylkGlDaW9k7cK1m", name="Makefile")

    # build the library
    os.system("make")

    from rcnn.processing.bbox_transform import clip_boxes
    from rcnn.processing.generate_anchor import generate_anchors_fpn, anchors_plane
    from rcnn.processing.nms import gpu_nms_wrapper, cpu_nms_wrapper


class RetinaFace:
    """Retina Face Model

    Args:
        config (config):
        nms (float):  Defaults to 0.4.
        nocrop (bool): Defaults to False.
        decay4 (float): Defaults to 0.5.
        vote (bool): Defaults to False.

    Examples:
        RetinaFace().detect_fast()

    """

    def __init__(self, config, nms=0.4, nocrop=False, decay4=0.5, vote=False):
        self.prefix = config["face_detection"]["path"]
        self.epoch = config["face_detection"]["epoch"]
        self.network = config["face_detection"]["network"]
        self.ctx_id = config["face_detection"]["gpuid"]
        self.decay4 = decay4
        self.nms_threshold = nms
        self.vote = vote
        self.nocrop = nocrop
        self.debug = False
        self.fpn_keys = []
        self.anchor_cfg = None
        pixel_means = [0.0, 0.0, 0.0]
        pixel_stds = [1.0, 1.0, 1.0]
        pixel_scale = 1.0
        self.preprocess = False
        _ratio = (1.0,)
        fmc = 3
        if self.network == "ssh" or self.network == "vgg":
            pixel_means = [103.939, 116.779, 123.68]
            self.preprocess = True
        elif self.network == "net3":
            _ratio = (1.0,)
        elif self.network == "net3a":
            _ratio = (1.0, 1.5)
        elif self.network == "net6":  # like pyramidbox or s3fd
            fmc = 6
        elif self.network == "net5":  # retinaface
            fmc = 5
        elif self.network == "net5a":
            fmc = 5
            _ratio = (1.0, 1.5)
        elif self.network == "net4":
            fmc = 4
        elif self.network == "net4a":
            fmc = 4
            _ratio = (1.0, 1.5)
        else:
            assert False, "network setting error %s" % self.network

        if fmc == 3:
            self._feat_stride_fpn = [32, 16, 8]
            self.anchor_cfg = {
                "32": {"SCALES": (32, 16), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "16": {"SCALES": (8, 4), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "8": {"SCALES": (2, 1), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
            }
        elif fmc == 4:
            self._feat_stride_fpn = [32, 16, 8, 4]
            self.anchor_cfg = {
                "32": {"SCALES": (32, 16), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "16": {"SCALES": (8, 4), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "8": {"SCALES": (2, 1), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "4": {"SCALES": (2, 1), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
            }
        elif fmc == 6:
            self._feat_stride_fpn = [128, 64, 32, 16, 8, 4]
            self.anchor_cfg = {
                "128": {"SCALES": (32,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "64": {"SCALES": (16,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "32": {"SCALES": (8,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "16": {"SCALES": (4,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "8": {"SCALES": (2,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
                "4": {"SCALES": (1,), "BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999},
            }
        elif fmc == 5:
            self._feat_stride_fpn = [64, 32, 16, 8, 4]
            self.anchor_cfg = {}
            _ass = 2.0 ** (1.0 / 3)
            _basescale = 1.0
            for _stride in [4, 8, 16, 32, 64]:
                key = str(_stride)
                value = {"BASE_SIZE": 16, "RATIOS": _ratio, "ALLOWED_BORDER": 9999}
                scales = []
                for _ in range(3):
                    scales.append(_basescale)
                    _basescale *= _ass
                value["SCALES"] = tuple(scales)
                self.anchor_cfg[key] = value

        for s in self._feat_stride_fpn:
            self.fpn_keys.append("stride%s" % s)

        dense_anchor = False
        self._anchors_fpn = dict(
            zip(self.fpn_keys, generate_anchors_fpn(dense_anchor=dense_anchor, cfg=self.anchor_cfg))
        )
        for k in self._anchors_fpn:
            v = self._anchors_fpn[k].astype(np.float32)
            self._anchors_fpn[k] = v

        self._num_anchors = dict(zip(self.fpn_keys, [anchors.shape[0] for anchors in self._anchors_fpn.values()]))
        sym, arg_params, aux_params = mx.model.load_checkpoint(self.prefix, self.epoch)

        if self.ctx_id >= 0:
            self.ctx = mx.gpu(self.ctx_id)
            self.nms = gpu_nms_wrapper(self.nms_threshold, self.ctx_id)
        else:
            self.ctx = mx.cpu()
            self.nms = cpu_nms_wrapper(self.nms_threshold)

        self.pixel_means = np.array(pixel_means, dtype=np.float32)
        self.pixel_stds = np.array(pixel_stds, dtype=np.float32)
        self.pixel_scale = float(pixel_scale)

        if len(sym) // len(self._feat_stride_fpn) == 3:
            self.use_landmarks = True

        if self.debug:
            c = len(sym) // len(self._feat_stride_fpn)
            sym = sym[(c * 0) :]
            self._feat_stride_fpn = [32, 16, 8]

        image_size = (640, 640)  # (406, 406 )
        self.model = mx.mod.Module(symbol=sym, context=self.ctx, label_names=None)
        self.model.bind(data_shapes=[("data", (1, 3, image_size[0], image_size[1]))], for_training=False)
        self.model.set_params(arg_params, aux_params)

        # buffer tensor
        self.buff_tensor = None
        self.buff_tensor_is_None = True

    def detect(self, img, threshold=0.5, scales=[1.0], do_flip=False):

        proposals_list = []
        scores_list = []
        landmarks_list = []
        strides_list = []
        timea = datetime.datetime.now()
        flips = [0]
        if do_flip:
            flips = [0, 1]

        imgs = [img]
        if isinstance(img, list):
            imgs = img
        for img in imgs:
            for im_scale in scales:
                for flip in flips:
                    if im_scale != 1.0:
                        im = cv2.resize(img, None, None, fx=im_scale, fy=im_scale, interpolation=cv2.INTER_LINEAR)
                    else:
                        im = img.copy()
                    if flip:
                        im = im[:, ::-1, :]
                    if self.nocrop:
                        if im.shape[0] % 32 == 0:
                            h = im.shape[0]
                        else:
                            h = (im.shape[0] // 32 + 1) * 32
                        if im.shape[1] % 32 == 0:
                            w = im.shape[1]
                        else:
                            w = (im.shape[1] // 32 + 1) * 32
                        _im = np.zeros((h, w, 3), dtype=np.float32)
                        _im[0 : im.shape[0], 0 : im.shape[1], :] = im
                        im = _im
                    else:
                        im = im.astype(np.float32)
                    if self.debug:
                        timeb = datetime.datetime.now()
                        diff = timeb - timea
                        print("X1 uses", diff.total_seconds(), "seconds")

                    im_info = [im.shape[0], im.shape[1]]
                    im_tensor = np.zeros((1, 3, im.shape[0], im.shape[1]))
                    for i in range(3):
                        im_tensor[0, i, :, :] = (
                            im[:, :, 2 - i] / self.pixel_scale - self.pixel_means[2 - i]
                        ) / self.pixel_stds[2 - i]
                    if self.debug:
                        timeb = datetime.datetime.now()
                        diff = timeb - timea
                        print("X2 uses", diff.total_seconds(), "seconds")
                    data = nd.array(im_tensor)
                    db = mx.io.DataBatch(data=(data,), provide_data=[("data", data.shape)])
                    if self.debug:
                        timeb = datetime.datetime.now()
                        diff = timeb - timea
                        print("X3 uses", diff.total_seconds(), "seconds")
                    self.model.forward(db, is_train=False)
                    net_out = self.model.get_outputs()

                    sym_idx = 0

                    for _idx, s in enumerate(self._feat_stride_fpn):

                        _key = "stride%s" % s
                        stride = int(s)
                        is_cascade = False
                        if self.cascade:
                            is_cascade = True

                        scores = net_out[sym_idx].asnumpy()
                        if self.debug:
                            timeb = datetime.datetime.now()
                            diff = timeb - timea
                            print("A uses", diff.total_seconds(), "seconds")

                        scores = scores[:, self._num_anchors["stride%s" % s] :, :, :]

                        bbox_deltas = net_out[sym_idx + 1].asnumpy()

                        height, width = bbox_deltas.shape[2], bbox_deltas.shape[3]

                        A = self._num_anchors["stride%s" % s]
                        K = height * width
                        anchors_fpn = self._anchors_fpn["stride%s" % s]
                        anchors = anchors_plane(height, width, stride, anchors_fpn)

                        anchors = anchors.reshape((K * A, 4))

                        scores = scores.transpose((0, 2, 3, 1)).reshape((-1, 1))

                        bbox_deltas = bbox_deltas.transpose((0, 2, 3, 1))
                        bbox_pred_len = bbox_deltas.shape[3] // A

                        bbox_deltas = bbox_deltas.reshape((-1, bbox_pred_len))
                        bbox_deltas[:, 0::4] = bbox_deltas[:, 0::4] * self.bbox_stds[0]
                        bbox_deltas[:, 1::4] = bbox_deltas[:, 1::4] * self.bbox_stds[1]
                        bbox_deltas[:, 2::4] = bbox_deltas[:, 2::4] * self.bbox_stds[2]
                        bbox_deltas[:, 3::4] = bbox_deltas[:, 3::4] * self.bbox_stds[3]
                        proposals = self.bbox_pred(anchors, bbox_deltas)

                        if is_cascade:
                            cascade_sym_num = 0
                            cls_cascade = False
                            bbox_cascade = False
                            __idx = [3, 4]
                            if not self.use_landmarks:
                                __idx = [2, 3]
                            for diff_idx in __idx:
                                if sym_idx + diff_idx >= len(net_out):
                                    break
                                body = net_out[sym_idx + diff_idx].asnumpy()
                                if body.shape[1] // A == 2:
                                    if cls_cascade or bbox_cascade:
                                        break
                                    else:
                                        cascade_scores = body[:, self._num_anchors["stride%s" % s] :, :, :]
                                        cascade_scores = cascade_scores.transpose((0, 2, 3, 1)).reshape((-1, 1))

                                        scores = cascade_scores
                                        cascade_sym_num += 1
                                        cls_cascade = True

                                elif body.shape[1] // A == 4:
                                    cascade_deltas = body.transpose((0, 2, 3, 1)).reshape((-1, bbox_pred_len))
                                    cascade_deltas[:, 0::4] = cascade_deltas[:, 0::4] * self.bbox_stds[0]
                                    cascade_deltas[:, 1::4] = cascade_deltas[:, 1::4] * self.bbox_stds[1]
                                    cascade_deltas[:, 2::4] = cascade_deltas[:, 2::4] * self.bbox_stds[2]
                                    cascade_deltas[:, 3::4] = cascade_deltas[:, 3::4] * self.bbox_stds[3]
                                    proposals = self.bbox_pred(proposals, cascade_deltas)
                                    cascade_sym_num += 1
                                    bbox_cascade = True

                        proposals = clip_boxes(proposals, im_info[:2])

                        if stride == 4 and self.decay4 < 1.0:
                            scores *= self.decay4

                        scores_ravel = scores.ravel()

                        order = np.where(scores_ravel >= threshold)[0]

                        proposals = proposals[order, :]
                        scores = scores[order]
                        if flip:
                            oldx1 = proposals[:, 0].copy()
                            oldx2 = proposals[:, 2].copy()
                            proposals[:, 0] = im.shape[1] - oldx2 - 1
                            proposals[:, 2] = im.shape[1] - oldx1 - 1

                        proposals[:, 0:4] /= im_scale

                        proposals_list.append(proposals)
                        scores_list.append(scores)
                        if self.nms_threshold < 0.0:
                            _strides = np.empty(shape=(scores.shape), dtype=np.float32)
                            _strides.fill(stride)
                            strides_list.append(_strides)

                        if not self.vote and self.use_landmarks:
                            landmark_deltas = net_out[sym_idx + 2].asnumpy()

                            landmark_pred_len = landmark_deltas.shape[1] // A
                            landmark_deltas = landmark_deltas.transpose((0, 2, 3, 1)).reshape(
                                (-1, 5, landmark_pred_len // 5)
                            )
                            landmark_deltas *= self.landmark_std

                            landmarks = self.landmark_pred(anchors, landmark_deltas)
                            landmarks = landmarks[order, :]

                            if flip:
                                landmarks[:, :, 0] = im.shape[1] - landmarks[:, :, 0] - 1

                                order = [1, 0, 2, 4, 3]
                                flandmarks = landmarks.copy()
                                for idx, a in enumerate(order):
                                    flandmarks[:, idx, :] = landmarks[:, a, :]

                                landmarks = flandmarks
                            landmarks[:, :, 0:2] /= im_scale

                            landmarks_list.append(landmarks)

                        if self.use_landmarks:
                            sym_idx += 3
                        else:
                            sym_idx += 2
                        if is_cascade:
                            sym_idx += cascade_sym_num

        if self.debug:
            timeb = datetime.datetime.now()
            diff = timeb - timea
            print("B uses", diff.total_seconds(), "seconds")
        proposals = np.vstack(proposals_list)
        landmarks = None
        if proposals.shape[0] == 0:
            if self.use_landmarks:
                landmarks = np.zeros((0, 5, 2))
            if self.nms_threshold < 0.0:
                return np.zeros((0, 6)), landmarks
            else:
                return np.zeros((0, 5)), landmarks
        scores = np.vstack(scores_list)

        scores_ravel = scores.ravel()
        order = scores_ravel.argsort()[::-1]

        proposals = proposals[order, :]
        scores = scores[order]
        if self.nms_threshold < 0.0:
            strides = np.vstack(strides_list)
            strides = strides[order]
        if not self.vote and self.use_landmarks:
            landmarks = np.vstack(landmarks_list)
            landmarks = landmarks[order].astype(np.float32, copy=False)

        if self.nms_threshold > 0.0:
            pre_det = np.hstack((proposals[:, 0:4], scores)).astype(np.float32, copy=False)
            if not self.vote:
                keep = self.nms(pre_det)
                det = np.hstack((pre_det, proposals[:, 4:]))
                det = det[keep, :]
                if self.use_landmarks:
                    landmarks = landmarks[keep]
            else:
                det = np.hstack((pre_det, proposals[:, 4:]))
                det = self.bbox_vote(det)
        elif self.nms_threshold < 0.0:
            det = np.hstack((proposals[:, 0:4], scores, strides)).astype(np.float32, copy=False)
        else:
            det = np.hstack((proposals[:, 0:4], scores)).astype(np.float32, copy=False)

        if self.debug:
            timeb = datetime.datetime.now()
            diff = timeb - timea
            print("C uses", diff.total_seconds(), "seconds")
        return det, landmarks

    def detect_fast(self, net_data, detect_im_shape, threshold=0.5, scales=[1.0], do_flip=False):
        """Detect Fast

        Args:
            net_data ():
            detect_im_shape ():
            threshold ():
            scales ():
            do_flip ():

        Returns:

        """

        im_info = detect_im_shape
        proposals_list = []
        scores_list = []
        landmarks_list = []
        flips = [0]
        if do_flip:
            flips = [0, 1]

        for im_scale in scales:
            for flip in flips:
                data = net_data

                db = mx.io.DataBatch(data=(data,), provide_data=[("data", data.shape)])

                self.model.forward(db, is_train=False)

                net_out = self.model.get_outputs()
                # nd.waitall()

                # ctx = mx.cpu()
                # net_out = []
                # for i in range(len(net_out_0)):
                #     net_out.append(net_out_0[i].as_in_context(ctx))

                cnt = 0
                for _idx, s in enumerate(self._feat_stride_fpn):

                    cnt += 1
                    _key = "stride%s" % s

                    stride = int(s)

                    if self.use_landmarks:
                        idx = _idx * 3
                    else:
                        idx = _idx * 2

                    scores = net_out[idx].asnumpy()

                    A = self._num_anchors[_key]

                    scores = scores[:, A:, :, :]

                    idx += 1
                    bbox_deltas = net_out[idx].asnumpy()

                    height, width = bbox_deltas.shape[2], bbox_deltas.shape[3]

                    K = height * width
                    anchors_fpn = self._anchors_fpn[_key]

                    anchors = anchors_plane(height, width, stride, anchors_fpn)
                    anchors = anchors.reshape((K * A, 4))

                    scores = self._clip_pad(scores, (height, width))
                    scores = scores.transpose((0, 2, 3, 1)).reshape((-1, 1))

                    bbox_deltas = self._clip_pad(bbox_deltas, (height, width))
                    bbox_deltas = bbox_deltas.transpose((0, 2, 3, 1))
                    bbox_pred_len = bbox_deltas.shape[3] // A
                    bbox_deltas = bbox_deltas.reshape((-1, bbox_pred_len))

                    proposals = self.bbox_pred(anchors, bbox_deltas)
                    proposals = clip_boxes(proposals, im_info[:2])

                    scores_ravel = scores.ravel()

                    order = np.where(scores_ravel >= threshold)[0]

                    proposals = proposals[order, :]
                    scores = scores[order]
                    if stride == 4 and self.decay4 < 1.0:
                        scores *= self.decay4
                    if flip:
                        oldx1 = proposals[:, 0].copy()
                        oldx2 = proposals[:, 2].copy()
                        proposals[:, 0] = im_info[1] - oldx2 - 1
                        proposals[:, 2] = im_info[1] - oldx1 - 1

                    proposals[:, 0:4] /= im_scale

                    proposals_list.append(proposals)
                    scores_list.append(scores)

                    if not self.vote and self.use_landmarks:
                        idx += 1
                        landmark_deltas = net_out[idx].asnumpy()
                        landmark_deltas = self._clip_pad(landmark_deltas, (height, width))

                        landmark_pred_len = landmark_deltas.shape[1] // A
                        landmark_deltas = landmark_deltas.transpose((0, 2, 3, 1)).reshape(
                            (-1, 5, landmark_pred_len // 5)
                        )
                        landmarks = self.landmark_pred(anchors, landmark_deltas)
                        landmarks = landmarks[order, :]

                        if flip:
                            landmarks[:, :, 0] = im_info[1] - landmarks[:, :, 0] - 1

                            order = [1, 0, 2, 4, 3]
                            flandmarks = landmarks.copy()
                            for idx, a in enumerate(order):
                                flandmarks[:, idx, :] = landmarks[:, a, :]
                            landmarks = flandmarks
                        landmarks[:, :, 0:2] /= im_scale

                        landmarks_list.append(landmarks)

        proposals = np.vstack(proposals_list)
        landmarks = None
        if proposals.shape[0] == 0:
            if self.use_landmarks:
                landmarks = np.zeros((0, 5, 2))
            return np.zeros((0, 5)), landmarks
        scores = np.vstack(scores_list)
        scores_ravel = scores.ravel()

        order = scores_ravel.argsort()[::-1]

        proposals = proposals[order, :]
        scores = scores[order]

        t2 = time.time()

        t1 = time.time()

        if not self.vote and self.use_landmarks:
            landmarks = np.vstack(landmarks_list)
            landmarks = landmarks[order].astype(np.float32, copy=False)

        pre_det = np.hstack((proposals[:, 0:4], scores)).astype(np.float32, copy=False)
        if not self.vote:
            keep = self.nms(pre_det)
            det = np.hstack((pre_det, proposals[:, 4:]))
            det = det[keep, :]
            if self.use_landmarks:
                landmarks = landmarks[keep]
        else:
            det = np.hstack((pre_det, proposals[:, 4:]))
            det = self.bbox_vote(det)
            
        self.ctx.empty_cache() 
        del net_data
        

        return det, landmarks

    @staticmethod
    def _clip_pad(tensor, pad_shape):
        """
        Clip boxes of the pad area.
        :param tensor: [n, c, H, W]
        :param pad_shape: [h, w]
        :return: [n, c, h, w]
        """
        H, W = tensor.shape[2:]
        h, w = pad_shape

        if h < H or w < W:
            tensor = tensor[:, :, :h, :w].copy()

        return tensor

    @staticmethod
    def bbox_pred(boxes, box_deltas):
        """
        Transform the set of class-agnostic boxes into class-specific boxes
        by applying the predicted offsets (box_deltas)
        :param boxes: !important [N 4]
        :param box_deltas: [N, 4 * num_classes]
        :return: [N 4 * num_classes]
        """
        if boxes.shape[0] == 0:
            return np.zeros((0, box_deltas.shape[1]))

        boxes = boxes.astype(np.float, copy=False)
        widths = boxes[:, 2] - boxes[:, 0] + 1.0
        heights = boxes[:, 3] - boxes[:, 1] + 1.0
        ctr_x = boxes[:, 0] + 0.5 * (widths - 1.0)
        ctr_y = boxes[:, 1] + 0.5 * (heights - 1.0)

        dx = box_deltas[:, 0:1]
        dy = box_deltas[:, 1:2]
        dw = box_deltas[:, 2:3]
        dh = box_deltas[:, 3:4]

        pred_ctr_x = dx * widths[:, np.newaxis] + ctr_x[:, np.newaxis]
        pred_ctr_y = dy * heights[:, np.newaxis] + ctr_y[:, np.newaxis]
        pred_w = np.exp(dw) * widths[:, np.newaxis]
        pred_h = np.exp(dh) * heights[:, np.newaxis]

        pred_boxes = np.zeros(box_deltas.shape)
        # x1
        pred_boxes[:, 0:1] = pred_ctr_x - 0.5 * (pred_w - 1.0)
        # y1
        pred_boxes[:, 1:2] = pred_ctr_y - 0.5 * (pred_h - 1.0)
        # x2
        pred_boxes[:, 2:3] = pred_ctr_x + 0.5 * (pred_w - 1.0)
        # y2
        pred_boxes[:, 3:4] = pred_ctr_y + 0.5 * (pred_h - 1.0)

        if box_deltas.shape[1] > 4:
            pred_boxes[:, 4:] = box_deltas[:, 4:]

        return pred_boxes

    @staticmethod
    def landmark_pred(boxes, landmark_deltas):
        """Landmark prediction

        Args:
            boxes ():
            landmark_deltas ():

        Returns:

        """

        if boxes.shape[0] == 0:
            return np.zeros((0, landmark_deltas.shape[1]))
        boxes = boxes.astype(np.float, copy=False)
        widths = boxes[:, 2] - boxes[:, 0] + 1.0
        heights = boxes[:, 3] - boxes[:, 1] + 1.0
        ctr_x = boxes[:, 0] + 0.5 * (widths - 1.0)
        ctr_y = boxes[:, 1] + 0.5 * (heights - 1.0)
        pred = landmark_deltas.copy()
        for i in range(5):
            pred[:, i, 0] = landmark_deltas[:, i, 0] * widths + ctr_x
            pred[:, i, 1] = landmark_deltas[:, i, 1] * heights + ctr_y
        return pred

    def bbox_vote(self, det):

        if det.shape[0] == 0:
            dets = np.array([[10, 10, 20, 20, 0.002]])
            det = np.empty(shape=[0, 5])
        while det.shape[0] > 0:
            # IOU
            area = (det[:, 2] - det[:, 0] + 1) * (det[:, 3] - det[:, 1] + 1)
            xx1 = np.maximum(det[0, 0], det[:, 0])
            yy1 = np.maximum(det[0, 1], det[:, 1])
            xx2 = np.minimum(det[0, 2], det[:, 2])
            yy2 = np.minimum(det[0, 3], det[:, 3])
            w = np.maximum(0.0, xx2 - xx1 + 1)
            h = np.maximum(0.0, yy2 - yy1 + 1)
            inter = w * h
            o = inter / (area[0] + area[:] - inter)

            # nms
            merge_index = np.where(o >= self.nms_threshold)[0]
            det_accu = det[merge_index, :]
            det = np.delete(det, merge_index, 0)
            if merge_index.shape[0] <= 1:
                if det.shape[0] == 0:
                    try:
                        dets = np.row_stack((dets, det_accu))
                    except:
                        dets = det_accu
                continue
            det_accu[:, 0:4] = det_accu[:, 0:4] * np.tile(det_accu[:, -1:], (1, 4))
            max_score = np.max(det_accu[:, 4])
            det_accu_sum = np.zeros((1, 5))
            det_accu_sum[:, 0:4] = np.sum(det_accu[:, 0:4], axis=0) / np.sum(det_accu[:, -1:])
            det_accu_sum[:, 4] = max_score
            try:
                dets = np.row_stack((dets, det_accu_sum))
            except:
                dets = det_accu_sum
        dets = dets[0:750, :]
        return dets




"""
Logger class
============

Wrapper of built-in `logging` Python module
that supports logging task to both console and files
"""

"""
Revision
--------
    2019, Sep 24: first version
"""


import os
import logging

# format for logging message
LOG_FILE_FORMAT = '%(asctime)s  %(name)12s  [%(levelname)s]  %(message)s'
LOG_CONSOLE_FORMAT = '%(asctime)s  [%(levelname)s]  %(message)s'


class Logger():
    """Logger class for logging task"""

    def __init__(self, name, console=True, set_level = 'DEBUG'):
    # def __init__(self, name, console=True, set_level = LEVELS['debug']):
        """Initialize logger

        Parameters
        ----------
        name : str
            name of logger and file to log
        console : bool
            whether logging messages are emitted to console or not
        set_level: str
            default level to log 
        """
        self.logger = logging.getLogger(name)
        
        self.__configLogger(name, console, set_level)

    
    def __configLogger(self, name, console, set_level):
        """Configure logger object

        Parameters
        ----------
        name : str
            name of logging file
        console : bool
            whether logging messages are emitted to console or not
        """
        self.logger.setLevel(set_level)
        file_formatter = logging.Formatter(LOG_FILE_FORMAT)
        console_formatter = logging.Formatter(LOG_CONSOLE_FORMAT)

        # if `console` is set, setup Handler to process 
        if console:
            console_log = logging.StreamHandler()
            console_log.setLevel(logging.DEBUG)
            console_log.setFormatter(console_formatter)

        file_log = logging.FileHandler(name + '.log',
                                      mode='w')
        file_log.setLevel(logging.DEBUG)
        file_log.setFormatter(file_formatter)

        if console:
            self.logger.addHandler(console_log)
        self.logger.addHandler(file_log)


    def critical(self, msg):
        """Emit CRITICAL message"""
        self.logger.critical(msg)


    def error(self, msg):
        """Emit ERROR message"""
        self.logger.error(msg)


    def warning(self, msg):
        """Emit WARNING message"""
        self.logger.warning(msg)


    def info(self, msg):
        """Emit INFO message"""
        self.logger.info(msg)


    def debug(self, msg):
        """Emit DEBUG message"""
        self.logger.debug(msg)


if __name__ == '__main__':
    LOG = Logger('app-face-detector',set_level = 'DEBUG')

    LOG.debug('message')
    LOG.info('message')
    LOG.warning('message')
    LOG.error('message')
    LOG.critical('message')


# import the necessary packages
from threading import Thread
import sys
import cv2
import time

# import the Queue class from Python 3
if sys.version_info >= (3, 0):
	from queue import Queue

# otherwise, import the Queue class for Python 2.7
else:
	from Queue import Queue


class FileVideoStream:
	def __init__(self, path, transform=None, queue_size=128):
		# initialize the file video stream along with the boolean
		# used to indicate if the thread should be stopped or not
		self.stream = cv2.VideoCapture(path)
		self.stopped = False
		self.transform = transform

		# initialize the queue used to store frames read from
		# the video file
		self.Q = Queue(maxsize=queue_size)
		# intialize thread
		self.thread = Thread(target=self.update, args=())
		self.thread.daemon = True

	def start(self):
		# start a thread to read frames from the file video stream
		self.thread.start()
		return self

	def update(self):
		# keep looping infinitely
		while True:
			# if the thread indicator variable is set, stop the
			# thread
			if self.stopped:
				break

			# otherwise, ensure the queue has room in it
			if not self.Q.full():
				# read the next frame from the file
				(grabbed, frame) = self.stream.read()

				# if the `grabbed` boolean is `False`, then we have
				# reached the end of the video file
				if not grabbed:
					self.stopped = True
					
				# if there are transforms to be done, might as well
				# do them on producer thread before handing back to
				# consumer thread. ie. Usually the producer is so far
				# ahead of consumer that we have time to spare.
				#
				# Python is not parallel but the transform operations
				# are usually OpenCV native so release the GIL.
				#
				# Really just trying to avoid spinning up additional
				# native threads and overheads of additional
				# producer/consumer queues since this one was generally
				# idle grabbing frames.
				if self.transform:
					frame = self.transform(frame)

				# add the frame to the queue
				self.Q.put(frame)
			else:
				time.sleep(0.1)  # Rest for 10ms, we have a full queue

		self.stream.release()

	def read(self):
		# return next frame in the queue
		return self.Q.get()

	# Insufficient to have consumer use while(more()) which does
	# not take into account if the producer has reached end of
	# file stream.
	def running(self):
		return self.more() or not self.stopped

	def more(self):
		# return True if there are still frames in the queue. If stream is not stopped, try to wait a moment
		tries = 0
		while self.Q.qsize() == 0 and not self.stopped and tries < 5:
			time.sleep(0.1)
			tries += 1

		return self.Q.qsize() > 0

	def stop(self):
		# indicate that the thread should be stopped
		self.stopped = True
		# wait until stream resources are released (producer thread might be still grabbing frame)
		self.thread.join()



import numpy as np
import cv2
import yaml
import gdown
import zipfile
import os
from shutil import copyfile


def unzip(src: str, dst: str = ""):
    """unzip the file

    Args:
        src (str):
        dst (str):

    Returns:
        None
    """
    with zipfile.ZipFile(src, 'r') as zip_ref:
        if len(dst):
            zip_ref.extractall(dst)
        else:
            zip_ref.extractall()


def get_cfg(path: str):
    """Set up the config
    
    Args:
        path (str):

    Returns:
        None
    """
    
    yaml.warnings({'YAMLLoadWarning': False})
    with open(path, 'r') as fs:
        config = yaml.load(fs)
    return config


def download_model(id: str, name: str, ext: str, dst: str = "./"):
    """Downloads file from a url to a destination.
    
    Args:
        id (str): url to download file.
        dst (str): destination path.
    """

    url = "https://drive.google.com/uc?id=" + id
    output = name + ext
    gdown.download(url, output, quiet=False)
    
    copyfile(output, os.path.join(dst,output))
    
    # clean up space
    os.remove(output)


def download_url(id: str, name: str, dst: str = "./"):
    """Downloads file from a url to a destination.
    
    Args:
        id (str): url to download file.
        dst (str): destination path.
    """

    url = "https://drive.google.com/uc?id=" + id
    output = name + '.zip'
#    time.sleep(100)
    gdown.download(url, output, quiet=False)

    # unzip the file
    unzip(output, dst)

    # clean up space
    os.remove(output)


def load_img(path):
    img = cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)
    img = img / 255.0
    img[:, :, 0] = (img[:, :, 0] - 0.485) / 0.229
    img[:, :, 1] = (img[:, :, 1] - 0.456) / 0.224
    img[:, :, 2] = (img[:, :, 2] - 0.406) / 0.225
    return img.astype(np.float32)

def preprocess_image(image, image_size):
    # image, RGB
    image_height, image_width = image.shape[:2]
    if image_height > image_width:
        scale = image_size / image_height
        resized_height = image_size
        resized_width = int(image_width * scale)
    else:
        scale = image_size / image_width
        resized_height = int(image_height * scale)
        resized_width = image_size

    image = cv2.resize(image, (resized_width, resized_height))
    image = image.astype(np.float32)
    image /= 255.
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]
    image -= mean
    image /= std
    pad_h = image_size - resized_height
    pad_w = image_size - resized_width
    image = np.pad(image, [(0, pad_h), (0, pad_w), (0, 0)], mode='constant')

    return image, scale


def load_processed_image(path, image_size=640):
    image = cv2.imread(path)

    # BGR -> RGB
    image = image[:, :, ::-1]
    h, w = image.shape[:2]
    image, scale = preprocess_image(image, image_size=image_size)
    return image, scale, h, w


def draw_boxes(image, boxes, scores, labels):
    num_classes = 90
    colors = [np.random.randint(0, 256, 3).tolist() for _ in range(num_classes)]
    for b, l, s in zip(boxes, labels, scores):
        class_id = int(l)
        class_name = 'person'
    
        xmin, ymin, xmax, ymax = list(map(int, b))
        score = '{:.4f}'.format(s)
        color = colors[class_id]
        label = '-'.join([class_name, score])
    
        ret, baseline = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
        cv2.rectangle(image, (xmin, ymin), (xmax, ymax), color, 1)
        cv2.rectangle(image, (xmin, ymax - ret[1] - baseline), (xmin + ret[0], ymax), color, -1)
        cv2.putText(image, label, (xmin, ymax - baseline), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 0), 1)



import statistics
from license_recognition.utils.tools import *

class OCR():
	""" OCR Model
	"""

	def __init__(self, config):
		self.model_ocr = None
		self.config = config
		
		self.ocr_classes = None

		self.load_classes()
		self.load_model_ocr()

	def load_classes(self):
		""" Load class names """

		model_classes = self.config['model2']['classes']

		with open(model_classes, 'rt') as f:
		    self.ocr_classes = f.read().rstrip('\n').split('\n')

	def load_model_ocr(self):
	    """Load orc model"""
	    
	    model_config = self.config['model2']['config']
	    model_weight = self.config['model2']['weight']
	    

	    self.model_ocr = cv2.dnn.readNetFromDarknet(model_config, model_weight)
	    self.model_ocr.setPreferableBackend(cv2.dnn.DNN_BACKEND_OPENCV)
	    self.model_ocr.setPreferableTarget(cv2.dnn.DNN_TARGET_CPU)


	def get_number(self, image):
	    """Get the string number from license plate

	    Args:
	        image (np.array): plate image after processing

	    Returns:
	        lp_str (str): number of plate
	    """        

	    H, W = image.shape[:2]

	    # Create a 4D blob from a image.
	    blob = cv2.dnn.blobFromImage(image, 1 / 255, (W, H), [0, 0, 0], 1, crop=False)

	    # Sets the input to the network
	    self.model_ocr.setInput(blob)

	    # Runs the forward pass to get output of the output layers
	    outs = self.model_ocr.forward(getOutputsNames(self.model_ocr))

	    # Remove the bounding boxes with low confidence
	    res = postprocess_ocr(image, outs, self.ocr_classes, 0.2, 0.1)

	    if len(res) <= 0:
	        return ""

	    L = dknet_label_conversion(res, W, H)
	    L = nms(L, .45)

	    # Compute the average height
	    heights = [l.wh()[1] for l in L]
	    avg_H = statistics.mean(heights)

	    # Sort letters by coordinate 'top'
	    L.sort(key=lambda x: x.tl()[1])

	    # Group the letters into rows
	    groups = do_grouping(L, bias=avg_H / 2)

	    # For each row, we sort letters from left to right
	    lp_str = ''
	    for g in groups:
	        g.sort(key=lambda x: x.tl()[0])
	        temp_str = ''.join([chr(l.cl()) for l in g])
	        lp_str = lp_str + temp_str + ' '

	    return lp_str


# author:    Adrian Rosebrock
# website:   http://www.pyimagesearch.com

# import the necessary packages
from __future__ import print_function
import cv2
import re

def find_function(name, pretty_print=True, module=None):
	# if the module is None, initialize it to to the root `cv2`
	# library
	if module is None:
		module = cv2

	# grab all function names that contain `name` from the module
	p = ".*{}.*".format(name)
	filtered = filter(lambda x: re.search(p, x, re.IGNORECASE), dir(module))
	
	# check to see if the filtered names should be returned to the
	# calling function
	if not pretty_print:
		return filtered

	# otherwise, loop over the function names and print them
	for (i, funcName) in enumerate(filtered):
		print("{}. {}".format(i + 1, funcName))


import numpy as np
import cv2
import tensorflow as tf
import yaml
from .label import Label
import gdown
import zipfile
import os

def unzip(src: str, dst: str = ""):
    """unzip the file

    Args:
        src (str):
        dst (str):

    Returns:
        None
    """
    with zipfile.ZipFile(src, 'r') as zip_ref:
        if len(dst):
            zip_ref.extractall(dst)
        else:
            zip_ref.extractall()


def get_cfg(path: str):
    """Set up the config
    
    Args:
        path (str):

    Returns:
        None
    """
    
    yaml.warnings({'YAMLLoadWarning': False})
    with open(path, 'r') as fs:
        config = yaml.load(fs)
    return config

def download_url(id: str, name: str, dst: str = ""):
    """Downloads file from a url to a destination.
    
    Args:
        id (str): url to download file.
        dst (str): destination path.
    """

    url = "https://drive.google.com/uc?id=" + id
    output = name + '.zip'
    gdown.download(url, output, quiet=False)

    # unzip the file
    unzip(output, dst)

    # clean up space
    os.remove(output)


class DLabel (Label):
    
    def __init__(self,cl,pts,prob):
        self.pts = pts
        tl = np.amin(pts,1)
        br = np.amax(pts,1)
        Label.__init__(self,cl,tl,br,prob)

def IOU(tl1,br1,tl2,br2):
    """[summary]

    Args:
        tl1 ([type]): [description]
        br1 ([type]): [description]
        tl2 ([type]): [description]
        br2 ([type]): [description]

    Returns:
        [type]: [description]
    """ 
   
    wh1,wh2 = br1-tl1,br2-tl2
    assert((wh1>=.0).all() and (wh2>=.0).all())
    
    intersection_wh = np.maximum(np.minimum(br1,br2) - np.maximum(tl1,tl2),0.)
    intersection_area = np.prod(intersection_wh)
    area1,area2 = (np.prod(wh1),np.prod(wh2))
    union_area = area1 + area2 - intersection_area
    return intersection_area/union_area

def get_width_height(ptsh):
    """[summary]

    Args:
        ptsh ([type]): [description]

    Returns:
        [type]: [description]
    """    
 
    x1,x2,x3,x4 = ptsh[0]
    y1,y2,y3,y4 = ptsh[1]
    width = (abs(x2-x1) + abs(x3-x4))/2
    height = (abs(y4-y1) + abs(y3-y2))/2

    return width, height

def find_T_matrix(pts,t_pts):
    """[summary]

    Args:
        pts ([type]): [description]
        t_pts ([type]): [description]

    Returns:
        [type]: [description]
    """
     
    A = np.zeros((8,9))
    for i in range(0,4):
        xi  = pts[:,i]
        xil = t_pts[:,i]
        xi  = xi.T
        
        A[i*2,   3:6] = -xil[2]*xi
        A[i*2,   6: ] =  xil[1]*xi
        A[i*2+1,  :3] =  xil[2]*xi
        A[i*2+1, 6: ] = -xil[0]*xi

    
    [U,S,V] = np.linalg.svd(A)
    H = V[-1,:].reshape((3,3))

    return H


def getRectPts(tlx,tly,brx,bry):
    return np.matrix([[tlx,brx,brx,tlx],[tly,tly,bry,bry],[1.,1.,1.,1.]],dtype=float)

def IOU_labels(l1,l2):
    return IOU(l1.tl(),l1.br(),l2.tl(),l2.br())

def getWH(shape):
    return np.array(shape[1::-1]).astype(float)

def nms(Labels,iou_threshold=.5):

    SelectedLabels = []
    Labels.sort(key=lambda l: l.prob(),reverse=True)
    
    for label in Labels:

        non_overlap = True
        for sel_label in SelectedLabels:
            if IOU_labels(label,sel_label) > iou_threshold:
                non_overlap = False
                break

        if non_overlap:
            SelectedLabels.append(label)

    return SelectedLabels

def im2single(img):
    """

    Args:
        img ():

    Returns:

    """

    assert(img.dtype == 'uint8')
    return img.astype('float32')/255.


def preprocess_img(img):
    """Prepare the input image for model

    Args:
        img (np.array): np.array image, HxWx3

    Returns:
        input_img (np.array): image after processing, 1xHxWx3
        resized_img (np.array): resized image with shape of HxWx3
        img (np.array): original image

    """

    net_step = 2**4

    ratio = float(max(img.shape[:2]))/min(img.shape[:2])
    side  = int(ratio*288.)

    # max dimension
    bound_dim = min(side + (side%(2**4)),608)

    single_img = im2single(img)

    min_dim_img = min(single_img.shape[:2])
    factor = float(bound_dim)/min_dim_img

    w,h = (np.array(single_img.shape[1::-1],dtype=float)*factor).astype(int).tolist()
    w += (w%net_step!=0)*(net_step - w%net_step)
    h += (h%net_step!=0)*(net_step - h%net_step)

    resized_img = cv2.resize(single_img,(w,h))

    input_img = resized_img.copy()
    input_img = input_img.reshape((1,input_img.shape[0],input_img.shape[1],input_img.shape[2]))

    return input_img,resized_img,img

def postprocess_plate(Iorig,I,Y):
    """

    Args:
        Iorig ():
        I ():
        Y ():

    Returns:

    """

    threshold = 0.9

    square_out_size = (160,160)
    rectangle_out_size = (240,80)

    net_stride  = 2**4
    side        = ((208. + 40.)/2.)/net_stride # 7.75

    Probs = Y[...,0]
    Affines = Y[...,2:]

    xx,yy = np.where(Probs>threshold)

    WH = getWH(I.shape)
    MN = WH/net_stride

    vxx = vyy = 0.5 #alpha

    base = lambda vx,vy: np.matrix([[-vx,-vy,1.],[vx,-vy,1.],[vx,vy,1.],[-vx,vy,1.]]).T
    labels = []

    for i in range(len(xx)):
        y,x = xx[i],yy[i]
        affine = Affines[y,x]
        prob = Probs[y,x]

        mn = np.array([float(x) + .5,float(y) + .5])

        A = np.reshape(affine,(2,3))
        A[0,0] = max(A[0,0],0.)
        A[1,1] = max(A[1,1],0.)

        pts = np.array(A*base(vxx,vyy)) #*alpha
        pts_MN_center_mn = pts*side
        pts_MN = pts_MN_center_mn + mn.reshape((2,1))

        pts_prop = pts_MN/MN.reshape((2,1))

        labels.append(DLabel(0,pts_prop,prob))

    final_labels = nms(labels,.1)
    TLps = []
    is_square_list = []

    if len(final_labels):
        final_labels.sort(key=lambda x: x.prob(), reverse=True)
        for i,label in enumerate(final_labels):
            ptsh    = np.concatenate((label.pts*getWH(Iorig.shape).reshape((2,1)),np.ones((1,4))))
            width, height = get_width_height(ptsh)

            is_square = True
            out_size = square_out_size

            if (width >= height * 2):
                out_size = rectangle_out_size
                is_square = False
            
            t_ptsh  = getRectPts(0,0,out_size[0],out_size[1])
            
            H       = find_T_matrix(ptsh,t_ptsh)
            Ilp     = cv2.warpPerspective(Iorig,H,out_size,borderValue=.0)

            TLps.append(Ilp)
            is_square_list.append(is_square)

    return final_labels,TLps,is_square_list

# Get the names of the output layers
def getOutputsNames(net):
    # Get the names of all the layers in the network
    layersNames = net.getLayerNames()
    # Get the names of the output layers, i.e. the layers with unconnected outputs
    return [layersNames[i[0] - 1] for i in net.getUnconnectedOutLayers()]


def postprocess_ocr(frame, outs,classes,confThreshold,nmsThreshold):
    """

    Args:
        frame ():
        outs ():
        classes ():
        confThreshold ():
        nmsThreshold ():

    Returns:

    """

    frameHeight = frame.shape[0]
    frameWidth = frame.shape[1]

    # Scan through all the bounding boxes output from the network and keep only the
    # ones with high confidence scores. Assign the box's class label as the class with the highest score.

    cnt = 0
    classIds = []
    confidences = []
    boxes = []

    for out in outs:
        for detection in out:
            scores = detection[5:]
            classId = np.argmax(scores)
            confidence = scores[classId]
            if (confidence > confThreshold):
                center_x = int(detection[0] * frameWidth)
                center_y = int(detection[1] * frameHeight)
                width = int(detection[2] * frameWidth)
                height = int(detection[3] * frameHeight)
                left = int(center_x - width / 2)
                top = int(center_y - height / 2)
                if (width >= 1) and (height >= 1):
                    classIds.append(classId)
                    confidences.append(float(confidence))
                    boxes.append([left, top, width, height])

                cnt += 1
    # print("cnt = ", cnt)

    res = []

    if cnt == 0:
        return res
    # Perform non maximum suppression to eliminate redundant overlapping boxes with
    # lower confidences.
    indices = cv2.dnn.NMSBoxes(boxes, confidences, confThreshold, nmsThreshold)

    for i in indices:
        i = i[0]
        box = boxes[i]
        res.append([classes[classIds[i]], confidences[i], box])

    return res

def dknet_label_conversion(R,img_width,img_height):
    WH = np.array([img_width,img_height],dtype=float)
    L  = []
    for r in R:
        center = np.array(r[2][:2])/WH
        wh2 = (np.array(r[2][2:])/WH)*.5
        L.append(Label(ord(r[0]),tl=center-wh2,br=center+wh2,prob=r[1]))
    return L

def do_grouping(labels,bias):
    prev = None
    group = []
    for l in labels:
        top = l.tl()[1]
        if not prev or abs(top - prev) <= bias:
            group.append(l)
        else:
            yield group
            group = [l]
        prev = top
    if group:
        yield group


# import the necessary packages
from .count_frames import count_frames
from .fps import FPS
from .videostream import VideoStream
from .webcamvideostream import WebcamVideoStream
from .filevideostream import FileVideoStream


import os
import tensorflow as tf
import numpy as np

from .EfficientDet.model import efficientdet
from .EfficientDet.utils import postprocess_boxes

class EfficientDet():
    """ Efficient Detection Model
    """

    def __init__(self, config):
        self.model = None
        self.config = config

        self.load_model()

    def load_model(self):
        """Load crowd counting model"""

        model_path = self.config['model2']['path']
        gpu_device = self.config['gpu']
        
        phi = self.config['model2']['params']['phi']
        weighted_bifpn = self.config['model2']['params']['weighted_bifpn']
        num_classes = self.config['model2']['params']['num_classes']
        score_threshold = self.config['model2']['params']['score_threshold']

        if gpu_device==-1:
            with tf.device('/cpu:0'):
                _, self.model = efficientdet(phi=phi,
                                            weighted_bifpn=weighted_bifpn,
                                            num_classes=num_classes,
                                            score_threshold=score_threshold)
                self.model.load_weights(model_path, by_name=True)
                print('Load the pretrained model sucessfully!')
        else:
            with tf.device('/gpu:{}'.format(gpu_device)):
                _, self.model = efficientdet(phi=phi,
                                            weighted_bifpn=weighted_bifpn,
                                            num_classes=num_classes,
                                            score_threshold=score_threshold)
                self.model.load_weights(model_path, by_name=True)
                print('Load the pretrained model sucessfully in GPU mode!')


    def get_predictions(self, image, scale, h, w):
        """ Get the predictions from the input image

        Args:
            image (np.array): the RGB image

        Returns:

        """
        try:
            # run network
            boxes, scores, labels = self.model.predict_on_batch([np.expand_dims(image, axis=0)])
            boxes, scores, labels = np.squeeze(boxes), np.squeeze(scores), np.squeeze(labels)
            boxes = postprocess_boxes(boxes=boxes, scale=scale, height=h, width=w)

            # select indices which have a score above the threshold
            score_threshold = self.config['model2']['params']['score_threshold']
            indices = np.where(scores[:] > score_threshold)[0]

            # select those detections
            boxes = boxes[indices]
            labels = labels[indices]

            # select indices which have a label 'person'
            indices = np.where(labels[:]==0)[0]

            # select those detections
            boxes = boxes[indices]
            labels = labels[indices]
            scores = scores[indices]
            return boxes, labels, scores
        except:
            return None, None, None



# import the necessary packages
from ..convenience import is_cv3
import cv2

def count_frames(path, override=False):
	# grab a pointer to the video file and initialize the total
	# number of frames read
	video = cv2.VideoCapture(path)
	total = 0

	# if the override flag is passed in, revert to the manual
	# method of counting frames
	if override:
		total = count_frames_manual(video)

	# otherwise, let's try the fast way first
	else:
		# lets try to determine the number of frames in a video
		# via video properties; this method can be very buggy
		# and might throw an error based on your OpenCV version
		# or may fail entirely based on your which video codecs
		# you have installed
		try:
			# check if we are using OpenCV 3
			if is_cv3():
				total = int(video.get(cv2.CAP_PROP_FRAME_COUNT))

			# otherwise, we are using OpenCV 2.4
			else:
				total = int(video.get(cv2.cv.CV_CAP_PROP_FRAME_COUNT))

		# uh-oh, we got an error -- revert to counting manually
		except:
			total = count_frames_manual(video)

	# release the video file pointer
	video.release()

	# return the total number of frames in the video
	return total

def count_frames_manual(video):
	# initialize the total number of frames read
	total = 0

	# loop over the frames of the video
	while True:
		# grab the current frame
		(grabbed, frame) = video.read()
	 
		# check to see if we have reached the end of the
		# video
		if not grabbed:
			break

		# increment the total number of frames read
		total += 1

	# return the total number of frames in the video file
	return total


# import the necessary packages
from .webcamvideostream import WebcamVideoStream

class VideoStream:
	def __init__(self, src=0, usePiCamera=False, resolution=(320, 240),
		framerate=32, **kwargs):
		# check to see if the picamera module should be used
		if usePiCamera:
			# only import the picamera packages unless we are
			# explicity told to do so -- this helps remove the
			# requirement of `picamera[array]` from desktops or
			# laptops that still want to use the `imutils` package
			from .pivideostream import PiVideoStream

			# initialize the picamera stream and allow the camera
			# sensor to warmup
			self.stream = PiVideoStream(resolution=resolution,
				framerate=framerate, **kwargs)

		# otherwise, we are using OpenCV so initialize the webcam
		# stream
		else:
			self.stream = WebcamVideoStream(src=src)

	def start(self):
		# start the threaded video stream
		return self.stream.start()

	def update(self):
		# grab the next frame from the stream
		self.stream.update()

	def read(self):
		# return the current frame
		return self.stream.read()

	def stop(self):
		# stop the thread and release any resources
		self.stream.stop()



from __future__ import absolute_import

import os
import sys

from .wpod import WPOD
from .ocr import OCR

from license_recognition.utils.tools import get_cfg, download_url

__model_factory = {
    # license plate detection
    'wpod': WPOD,

    # number extracting
    'ocr': OCR
}

_model_urls = {
    'wpod': '1xKNbPIERGA0vf3Y4vdENxvij2H1RXL6H',
    'ocr': '1uTsHKHnxH5fqfOGvqtrkw1OQU3Ue9Lw6'
}

def show_avai_models():
    """Displays available models.

    Examples::
        >>> from face_recognition import models
        >>> models.show_avai_models()
    """
    print(list(__model_factory.keys()))


def build_model(name: str, cfg_path: str = ""):
    """A function wrapper for building a model.

    Args:
        name (str): model name
        cfg_path (str): path to config file
    Returns:
        model instance.
    """
    avai_models = list(__model_factory.keys())
    if name not in avai_models:
        raise KeyError(
            'Unknown model: {}. Must be one of {}'.format(name, avai_models)
        )
    if not len(cfg_path):
        raise RuntimeError(
            'Please input the path config file'
        )
    # Load config file
    config = get_cfg(cfg_path)

    try:
        os.makedirs('models')
    except:
        pass

    # Check the models are existed or not
    if not os.path.exists('models/LicenseRecognition/ocr'):
        download_url(_model_urls['ocr'], name='ocr', dst = "models")

    if not os.path.exists('models/LicenseRecognition/wpod'):
        download_url(_model_urls['wpod'],name = 'wpod', dst = "models")

    return __model_factory[name](config)



from keras.models import model_from_json
from license_recognition.utils.tools import *

class WPOD():
    """ WPOD Model
    """

    def __init__(self, config):
        self.model_plate = None
        self.config = config

        self.load_model_plate()

    def load_model_plate(self):
        """Load plate detection model"""

        model_path = self.config['model1']['path']
        with tf.device('/cpu:0'):
            model_path = os.path.splitext(model_path)[0]
            with open('%s.json' % model_path, 'r') as json_file:
                model_json = json_file.read()

            self.model_plate = model_from_json(model_json, custom_objects={})
            self.model_plate.load_weights('%s.h5' % model_path)

    def get_plate(self, input, resized, origin):
        """ Get the license plate from image

        Args:
            input (np.array): the image
            resized (np.array):
            origin (np.array):

        Returns:

        """
        # sess = tf.compat.v1.Session()
        # graph = tf.compat.v1.get_default_graph()
        # with sess.as_default():
        #     with graph.as_default():
        output = self.model_plate.predict(input)
        output = np.squeeze(output)

        Llp, LlpImgs, is_square_list = postprocess_plate(origin, resized, output)

        import copy
        if len(LlpImgs):
            Ilp = LlpImgs[0]
            Ilp_ori = copy.deepcopy(Ilp)
            Ilp = cv2.cvtColor(Ilp, cv2.COLOR_BGR2GRAY)
            Ilp = cv2.cvtColor(Ilp,cv2.COLOR_GRAY2BGR)
            # why?
            res_img = (Ilp * 255.).astype(np.uint8)
            # res_img = Ilp
            is_square = is_square_list[0]
            LlpImgs.pop(0)

            return res_img, is_square,Ilp_ori

        return None, None,None